#!/usr/bin/env python

"""
Handlers for a person for whom scheduling is performed.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.content import Handler
from imiptools.data import get_uri
from imiptools.dates import format_datetime
from imiptools.handlers.common import CommonFreebusy
from imiptools.period import replace_overlapping
from imiptools.profile import Preferences

class PersonHandler(Handler):

    "Handling mechanisms specific to people."

    def _record(self, from_organiser=True, queue=False, cancel=False):

        oa = self.require_organiser_and_attendees(from_organiser)
        if not oa:
            return False

        (organiser, organiser_attr), attendees = organiser_item, attendees = oa

        # Handle notifications and invitations.

        if from_organiser:

            # Process each attendee separately.

            for attendee, attendee_attr in attendees.items():

                if not self.have_new_object(attendee):
                    continue

                # Set the complete event if not an additional occurrence.

                event = self.obj.to_node()
                recurrenceid = self.obj.get_value("RECURRENCE-ID")

                if not recurrenceid:
                    self.store.set_event(attendee, self.uid, event)
                else:
                    self.store.set_recurrence(attendee, self.uid, recurrenceid, event)

                # Queue any request.

                if queue:
                    self.store.queue_request(attendee, self.uid)
                elif cancel:
                    self.store.cancel_event(attendee, self.uid)

                    # No return message will occur to update the free/busy
                    # information, so this is done here.

                    freebusy = self.store.get_freebusy(attendee)
                    self.remove_from_freebusy(freebusy, attendee)

                    if self.publisher:
                        self.publisher.set_freebusy(attendee, freebusy)

                self.update_freebusy_from_organiser(attendee, organiser_item)

        # As organiser, update attendance.

        else:
            if self.merge_attendance(attendees, organiser):
                self.update_freebusy_from_attendees(organiser, attendees)

        return True

    def _record_freebusy(self, from_organiser=True):

        "Record free/busy information for the received information."

        if from_organiser:
            organiser_item = self.require_organiser(from_organiser)
            if not organiser_item:
                return

            senders = [organiser_item]
        else:
            oa = self.require_organiser_and_attendees(from_organiser)
            if not oa:
                return

            organiser_item, attendees = oa
            senders = attendees.items()

            if not senders:
                return

        freebusy = []

        for value in self.obj.get_values("FREEBUSY") or []:
            if not isinstance(value, list):
                value = [value]
            for v in value:
                try:
                    start, end = v.split("/", 1)
                    freebusy.append((start, end))
                except ValueError:
                    pass

        dtstart = format_datetime(self.obj.get_utc_datetime("DTSTART"))
        dtend = format_datetime(self.obj.get_utc_datetime("DTEND"))
        user = get_uri(self.recipient)

        for sender, sender_attr in senders:
            stored_freebusy = self.store.get_freebusy_for_other(user, sender)
            replace_overlapping(stored_freebusy, (dtstart, dtend), freebusy)
            self.store.set_freebusy_for_other(user, sender, stored_freebusy)

class Event(PersonHandler):

    "An event handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("An addition to an event has been received.", link=False)

    def cancel(self):

        "Queue a cancellation of any active event."

        self._record(from_organiser=True, queue=False, cancel=True)
        return self.wrap("A cancellation has been received.", link=False)

    def counter(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("A counter proposal has been received.", link=False)

    def declinecounter(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("A declining counter proposal has been received.", link=False)

    def publish(self):

        "Register details of any relevant event."

        self._record(from_organiser=True, queue=False)
        return self.wrap("Details of an event have been received.")

    def refresh(self):

        "Update details of any active event."

        self._record(from_organiser=True, queue=False)
        return self.wrap("An event update has been received.")

    def reply(self):

        "Record replies and notify the recipient."

        self._record(from_organiser=False, queue=False)
        return self.wrap("A reply has been received.")

    def request(self):

        "Hold requests and notify the recipient."

        self._record(from_organiser=True, queue=True)
        return self.wrap("A request has been received.")

class Freebusy(PersonHandler, CommonFreebusy):

    "A free/busy handler."

    def publish(self):

        "Register free/busy information."

        self._record_freebusy(from_organiser=True)

        # Produce a message if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_messages") == "notify":
            return self.wrap("A free/busy update has been received.", link=False)

    def reply(self):

        "Record replies and notify the recipient."

        self._record_freebusy(from_organiser=False)

        # Produce a message if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_messages") == "notify":
            return self.wrap("A reply to a free/busy request has been received.", link=False)

    def request(self):

        """
        Respond to a request by preparing a reply containing free/busy
        information for each indicated attendee.
        """

        # Produce a reply if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_sharing") == "share":
            return CommonFreebusy.request(self)

class Journal(PersonHandler):

    "A journal entry handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active entry.

        return self.wrap("An addition to a journal entry has been received.", link=False)

    def cancel(self):

        # NOTE: Queue a suggested modification to any active entry.

        return self.wrap("A cancellation has been received.", link=False)

    def publish(self):

        # NOTE: Register details of any relevant entry.

        self._record(from_organiser=True, queue=False)
        return self.wrap("Details of a journal entry have been received.")

class Todo(PersonHandler):

    "A to-do item handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active item.

        return self.wrap("An addition to an item has been received.", link=False)

    def cancel(self):

        # NOTE: Queue a suggested modification to any active item.

        return self.wrap("A cancellation has been received.", link=False)

    def counter(self):

        # NOTE: Queue a suggested modification to any active item.

        return self.wrap("A counter proposal has been received.", link=False)

    def declinecounter(self):

        # NOTE: Queue a suggested modification to any active item.

        return self.wrap("A declining counter proposal has been received.", link=False)

    def publish(self):

        "Register details of any relevant item."

        self._record(from_organiser=True, queue=False)
        return self.wrap("Details of an item have been received.")

    def refresh(self):

        "Update details of any active item."

        self._record(from_organiser=True, queue=False)
        return self.wrap("An item update has been received.")

    def reply(self):

        "Record replies and notify the recipient."

        self._record(from_organiser=False, queue=False)
        return self.wrap("A reply has been received.")

    def request(self):

        "Hold requests and notify the recipient."

        self._record(from_organiser=True, queue=True)
        return self.wrap("A request has been received.")

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

# vim: tabstop=4 expandtab shiftwidth=4
